# -*- coding: utf-8 -*-

import os
import sys
import ldap3
import logging
import logging.handlers
import tempfile
import re
import shutil
import time
import atexit
import signal
import pwd
import filecmp
import random
import importlib.util as importlib_util
from pprint import pformat
from optparse import OptionParser, OptionGroup, BadOptionError, SUPPRESS_HELP

from . import __version__

LDAP_SCOPES = {"base": ldap3.BASE, "onelevel": ldap3.LEVEL, "subtree": ldap3.SUBTREE}

USERNAME_POLICIES = ["SITE_ONLY", "SITE_FIRST", "GLOBAL_ONLY", "GLOBAL_FIRST"]
SUBJECT_TRANSFORMS = ["AUTO", "NONE", "FORCE"]


def scope_name(scope):
    return next(
        (k for k, v in LDAP_SCOPES.items() if v == scope),
        "Unrecognized scope: {}".format(scope),
    )


def as_list(obj):
    return obj if isinstance(obj, list) else [obj]


def order_preserve_uniq(seq):
    """
    """
    checked = []
    for e in seq:
        if e not in checked:
            checked.append(e)
    return checked


def attr_join(arg, source=None, add_default=True):
    if source and (source != "default" or add_default):
        return "_".join([arg, source])
    return arg


def transform_subjects(orig_subjects, mode):
    if mode == "NONE":
        return orig_subjects

    subjects = []
    for subject in orig_subjects:
        # Simplest assumption about the X509 format
        if mode == "AUTO" and subject.startswith("/"):
            subjects.append(subject)
            continue

        try:
            subjects.append(
                "/"
                + "/".join(
                    (lambda key, value: "%s=%s" % (key.upper(), value))(
                        *part.split("=")
                    )
                    for part in reversed(subject.split(","))
                )
            )
        except TypeError:
            logging.error("Failed to transform certificate subject {}".format(subject))
            if mode == "AUTO":
                subjects.append(subject)

    return subjects


class PassThroughOptionParser(OptionParser):
    def _process_long_opt(self, rargs, values):
        try:
            OptionParser._process_long_opt(self, rargs, values)
        except BadOptionError as err:
            if getattr(self, "raise_unknown", None):
                raise err
            else:
                self.largs.append(err.opt_str)

    def _process_short_opts(self, rargs, values):
        try:
            OptionParser._process_short_opts(self, rargs, values)
        except BadOptionError as err:
            if getattr(self, "raise_unknown", None):
                raise err
            else:
                self.largs.append(err.opt_str)


class LDAPSourceConnectionError(RuntimeError):
    pass


class Mapping(dict):
    """
    """

    def __init__(self):
        """
        """
        super(Mapping, self).__init__()

    def persist_to_file(
        self, path, logger, sort=None, no_local_accounts=False, dry_run=False
    ):
        """
        """
        mapfile = []
        for subject in self:
            all_logins = self[subject]
            all_logins = order_preserve_uniq(all_logins)
            login_string = ",".join(all_logins)
            if no_local_accounts:
                line = '"%s"' % subject
            elif not login_string:
                line = '"%s"' % subject
            else:
                line = '"%s" %s' % (subject, login_string)
            mapfile.append(line)

        mapfile.sort(key=sort)

        with tempfile.NamedTemporaryFile(mode="w", delete=True) as f:
            for line in mapfile:
                print(line, file=f)
            f.flush()

            # open the file again, read it in and parse it
            # to make sure it is acceptable
            with open(f.name, "r") as fcheck:
                count = 0
                for line in fcheck:
                    if line:
                        count += 1
                        m = re.search(r"^\".+\"\s*.*$", line.strip())
                        if not m:
                            msg = "Error parsing temporary file"
                            raise RuntimeError(msg)
                if count != len(mapfile):
                    msg = "Length of temporary file is not correct"
                    raise RuntimeError(msg)

            # copy the file into place
            if not dry_run:
                # this comparison avoids an OSError through short-circuit or
                mapping_differs = not os.path.isfile(path) or not filecmp.cmp(
                    f.name, path
                )

                try:
                    if mapping_differs:
                        shutil.copyfile(f.name, path)
                    else:
                        logger.info("No changes so not over writing %s" % path)
                except IOError as err:
                    msg = "Failed to write to file %s. %s" % (path, err.args)
                    raise RuntimeError(msg)
            else:
                logger.info("DRY_RUN is True so not writing output")


class FileMapping(Mapping):
    """
    """

    def __init__(self, path):
        """
        """
        super(FileMapping, self).__init__()
        self.path = path
        self.pattern = re.compile(r'^"(.+)"\s*(.+)*$')
        self.comment = re.compile(r"^\s*(#+|$)")
        self.parsed = False

    def parse_line(self, number, line):
        """
        """
        m = self.pattern.search(line.strip())
        if not m:
            msg = "Error parsing %s: cannot parse line %d: %s" % (
                self.path,
                number + 1,
                line,
            )
            raise RuntimeError(msg)

        subject = m.group(1)

        if m.group(2):
            login_string = m.group(2)
            logins = [s.strip() for s in login_string.split(",")]
        else:
            logins = []

        if subject in self:
            self[subject].extend(logins)
        else:
            self[subject] = logins

    def parse(self):
        """
        """
        with open(self.path, "r") as f:
            for number, line in enumerate(f):
                if self.comment.search(line):
                    continue

                self.parse_line(number, line)
        self.parsed = True


class ExistingMapping(FileMapping):
    """
    """

    def __init__(self, path, required=True):
        super(ExistingMapping, self).__init__(path)

        if required or os.path.isfile(path):
            self.parse()


class BlackListMapping(ExistingMapping):
    """
    """

    def __init__(self, path):
        """
        """
        super(BlackListMapping, self).__init__(path)

    def parse_line(self, number, line):
        """
        """
        m = re.search(r'^"(.+)"\s*(.+)*$', line.strip())
        if m:
            subject = m.group(1)
            if m.group(2):
                login_string = m.group(2)
            else:
                login_string = ""
        else:
            subject = str(number)
            login_string = line.strip()

        logins = [s.strip() for s in login_string.split(",")]

        self[subject] = logins


class LDAPMapping(dict):
    """
    """

    def __init__(self, manager, source):
        """
        """
        super(LDAPMapping, self).__init__()
        self.manager = manager
        self.source = source

        self.ldap_params = {}
        self.prepare_ldap_params()

        overrides = self.manager.get_config("OVERRIDES", source)
        if overrides:
            self.compile_regexps(overrides)

    def refresh(self):
        result = self.fetch_result()
        self.clear()
        if result:
            self.process_result(result)

    def compile_regexps(self, overrides):
        # Compile regexps for use later
        for collections in [
            list(rule.get("conditions", {}).values()) for rule in overrides
        ]:
            for conditions in collections:
                for condition in conditions:
                    if condition.startswith("!~"):
                        regexp = condition[2:]
                        if regexp not in self.regexps:
                            self.regexps[regexp] = re.compile(regexp)

                    elif condition.startswith("~"):
                        regexp = condition[1:]
                        if regexp not in self.regexps:
                            self.regexps[regexp] = re.compile(regexp)

    def prepare_ldap_params(self):
        manager = self.manager
        logger = manager.logger
        source = self.source

        # prepare search filter using configured groups and
        # membership attribute
        myfilter = ""
        groups = manager.get_config("GROUPS", source)
        for sublist in groups:
            subfilter = "".join(
                [
                    "(%s=%s)" % (manager.get_config("MEMBER_ATTR", source), group)
                    for group in sublist
                ]
            )
            if len(sublist) > 1:
                myfilter += "(&%s)" % (subfilter,)
            else:
                myfilter += "%s" % (subfilter,)

        if len(groups) > 1:
            myfilter = "(|%s)" % (myfilter)

        if manager.get_config("FILTER", source):
            if manager.get_config("FILTER", source).startswith("("):
                myfilter += "%s" % (manager.get_config("FILTER", source),)
            else:
                myfilter += "(%s)" % (manager.get_config("FILTER", source),)

        # we also require two attributes and add custom filter
        myfilter = "(&%s(%s=*)(%s=*))" % (
            myfilter,
            manager.get_config("SUBJECT_ATTR", source),
            manager.get_config("LOGIN_ATTR", source),
        )
        logger.debug("Using filter %s" % myfilter)

        # only return these attributes
        attrs = [
            manager.get_config("SUBJECT_ATTR", source),
            manager.get_config("LOGIN_ATTR", source),
        ]

        if manager.get_config("LOGIN_GROUPS_MAP", source):
            attrs.append(manager.get_config("MEMBER_ATTR", source))

        self.regexps = {}
        # If have local overrides configured
        overrides = manager.get_config("OVERRIDES", source)
        if overrides:
            logger.info("Overriding logins using mapping: %s" % (overrides,))
            # Get attributes required for LDAP query
            for extraattrs in [
                list(rule.get("conditions", {}).keys()) for rule in overrides
            ]:
                attrs.extend(extraattrs)
            attrs = list(set(attrs).difference(set(["_logins", "_subjects"])))

        self.ldap_params[source] = {"attrs": attrs, "filter": myfilter}

    def fetch_result(self):
        manager = self.manager
        logger = manager.logger
        source = self.source
        ldap_params = self.ldap_params[source]

        # open connection to LDAP server
        connected = False
        for url in manager.get_config("LDAP_URLS", source):
            logger.info("Connecting to %s ..." % url)

            try:
                connection = ldap3.Connection(
                    url, auto_bind=True, raise_exceptions=True
                )

            except ldap3.core.exceptions.LDAPException as exc:
                logger.warn("Unable to connect to %s: %s" % (url, str(exc)))
                continue

            connected = True
            logger.info("Successfully connected to %s" % url)
            break

        if not connected:
            msg = "Unable to connect to any LDAP server for source '{}'".format(source)
            raise LDAPSourceConnectionError(msg)

        # query LDAP
        base = manager.get_config("LDAP_BASE", source)
        scope = LDAP_SCOPES[manager.get_config("LDAP_SCOPE", source)]
        logger.debug("Base is %s" % base)
        logger.debug(
            "Scope is %s (%s)" % (manager.get_config("LDAP_SCOPE", source), scope)
        )
        logger.debug("Attrs are %s" % ldap_params["attrs"])
        try:
            connection.search(
                base, ldap_params["filter"], scope, attributes=ldap_params["attrs"]
            )
            logger.debug("Successfully queryied %s" % url)

        except ldap3.core.exceptions.LDAPException as e:
            logger.error("Error querying server %s: %s" % (url, e))
            logger.error("Base was %s" % base)
            logger.error(
                "Scope was %s (%s)" % (manager.get_config("LDAP_SCOPE", source), scope)
            )
            logger.error(
                "Filter was %s"
                % ldap_params["filter"].replace("(", "(\n").replace(")", "\n)")
            )
            raise RuntimeError()

        try:
            entries = [(e.entry_dn, e.entry_attributes_as_dict) for
                       e in connection.entries]
        except AttributeError:  # ldap3 < 2.0.0
            entries = [(e.entry_get_dn(), e.entry_get_attributes_dict()) for
                       e in connection.entries]

        # disconnect from LDAP
        try:
            connection.unbind()
            logger.debug("Successful unbind from %s" % url)
        except ldap3.LDAPException as exc:
            logger.warn("Error unbinding to %s: %s" % (url, exc))

        return entries

    def process_result(self, result):
        manager = self.manager
        logger = manager.logger
        source = self.source

        logger.info("Found %d users in LDAP" % len(result))

        attrNames = [manager.get_config("LOGIN_ATTR", source)]
        if manager.get_config("SITE", source):
            site_attribute = manager.get_config("SITE_ATTR", source).format(
                site=manager.get_config("SITE", source)
            )

            login_attr_site = "%s;%s" % (
                manager.get_config("LOGIN_ATTR", source),
                site_attribute,
            )
            if manager.get_config("USERNAME_POLICY", source) in [
                "SITE_ONLY",
                "SITE_FIRST",
            ]:
                attrNames.insert(0, login_attr_site)
            else:
                attrNames.append(login_attr_site)

        subject_attr = manager.get_config("SUBJECT_ATTR", source)
        subject_transform = manager.get_config("SUBJECT_TRANSFORM", source)

        # loop over each DN, extract certificate subjects
        # and logins and prepare mappings
        mappings = {}
        for dn, attrDict in result:
            logger.debug("%s has attributes %s" % (dn, str(attrDict)))

            subjects = []
            # Handle options described in VOPerson schema; https://github.com/voperson/voperson/blob/master/voPerson.md
            for attr, value in attrDict.items():
                if ";" in attr:
                    attr_split = attr.split(";")
                    attr_name = attr_split[0]
                    attr_options = attr_split[1:]
                else:
                    attr_name = attr
                    attr_options = []

                if attr_name == subject_attr and "prior" not in attr_options:
                    subjects.extend(value)

            if subject_transform != "NONE":
                subjects = transform_subjects(subjects, subject_transform)

            logins = []
            # the break statement below causes the first of the attrNames
            # that is found in the LDAP query to override the values in
            # later attrNames that may also be present in the query
            # By searching for the site-specific override value first, this
            # enables specifying overrides for only sites that need them.
            for name in attrNames:
                if name in attrDict:
                    for loginString in attrDict[name]:
                        accounts = loginString.split(",")
                        logins.extend(accounts)
                    if manager.get_config("USERNAME_POLICY", source) in [
                        "SITE_ONLY",
                        "GLOBAL_ONLY",
                    ]:
                        break

            for accounts, groups in (
                manager.get_config("LOGIN_GROUPS_MAP", source) or []
            ):
                for group in groups:
                    if group not in attrDict[manager.get_config("MEMBER_ATTR", source)]:
                        break
                else:
                    logins.extend(as_list(accounts))

            overrides = manager.get_config("OVERRIDES", source)
            if overrides:
                subjects, logins = self.apply_overrides(
                    overrides, subjects, logins, attrDict
                )

            for s in subjects:
                if s in self:
                    logger.warn("Subject %s is mapped to more than one person" % s)
                    logger.warn("Skipping subject %s" % s)
                else:
                    mappings[s] = logins

        # If login whitelist exists, then apply it
        if manager.get_config("LOGIN_WHITELIST", source):
            logger.debug(
                "Filtering logins using whitelist: %s"
                % (manager.get_config("LOGIN_WHITELIST", source),)
            )
            for subject, logins in list(mappings.items()):
                mappings[subject] = list(
                    set(manager.get_config("LOGIN_WHITELIST", source)) & set(logins)
                )

        self.update(mappings)

    def apply_overrides(self, overrides, subjects, logins, attrDict):
        # Apply each collection of rules in turn
        for rule in overrides:
            matches = []
            satisfied = True
            # Step trhough each attribute and its conditions
            for attribute, conditions in list(rule.get("conditions", {}).items()):
                # Get the attribute to be tested
                if attribute == "_logins":
                    values = logins
                elif attribute == "_subjects":
                    values = subjects
                else:
                    values = attrDict.get(attribute, [])

                # Work through conditions and if condition not satisfied then break early
                for condition in conditions:
                    # Negative regexp
                    if condition.startswith("!~"):
                        if any(
                            [
                                self.regexps[condition[2:]].match(value)
                                for value in values
                            ]
                        ):
                            satisfied = False
                            break

                    # Regexp
                    elif condition.startswith("~"):
                        mm = [
                            self.regexps[condition[1:]].match(value) for value in values
                        ]
                        if any(mm):
                            [matches.extend(m.groups()) for m in mm if m]
                        else:
                            satisfied = False
                            break

                    # Does not contain
                    elif condition.startswith("!"):
                        if condition[1:] in values:
                            satisfied = False
                            break

                    # Contains
                    else:
                        if condition in values:
                            matches.append(condition)
                        else:
                            satisfied = False
                            break

            # If every condition of every attribute satisfied
            if satisfied:
                format_strings = dict(
                    [(k, ",".join(v)) for k, v in list(attrDict.items())]
                )

                # Apply changes to logins
                if rule.get("logins_set") is not None:
                    logins = [
                        log.format(_matches=matches, **format_strings)
                        for log in rule.get("logins_set", [])
                    ]

                logins_prepend = [
                    log.format(_matches=matches, **format_strings)
                    for log in rule.get("logins_prepend", [])
                ]
                logins_append = [
                    log.format(_matches=matches, **format_strings)
                    for log in rule.get("logins_append", [])
                ]
                logins = logins_prepend + logins + logins_append

                for item in rule.get("logins_remove", []):
                    if isinstance(item, int) and len(logins) > item:
                        logins.pop(item)
                    else:
                        item = item.format(_matches=matches, **format_strings)
                        while item in logins:
                            logins.remove(item)

                # Apply changes to subjects
                if rule.get("subjects_set") is not None:
                    subjects = [
                        sub.format(_matches=matches, **format_strings)
                        for sub in rule.get("subjects_set", [])
                    ]
                subjects_add = [
                    sub.format(_matches=matches, **format_strings)
                    for sub in rule.get("subjects_add", [])
                ]
                subjects += subjects_add

                for item in rule.get("subjects_remove", []):
                    if isinstance(item, int) and len(subjects) > item:
                        subjects.pop(item)
                    else:
                        item = item.format(_matches=matches, **format_strings)
                        while item in subjects:
                            subjects.remove(item)

        return subjects, logins


class CentralMapping(Mapping):
    """
    """

    def __init__(self, manager, fetch_now=True):
        """
        """
        super(CentralMapping, self).__init__()
        self.manager = manager
        self.mappings = {}
        for source in self.manager.SOURCES:
            self.mappings[source] = LDAPMapping(self.manager, source)

        if fetch_now:
            self.refresh()

    def refresh(self):
        logger = self.manager.logger
        self.clear()

        for source in self.manager.SOURCES:
            mapping = self.mappings[source]

            try:
                mapping.refresh()
            except LDAPSourceConnectionError as err:
                if self.manager.DAEMON:
                    logger.warn(err.msg)
                else:
                    raise RuntimeError(err.msg)

            for subject, logins in mapping.items():
                if subject in self:
                    logger.warn(
                        "Subject {} exists in more than one source".format(subject)
                    )
                    logger.warn(
                        "Skipping subject {} in source {}".format(subject, source)
                    )
                else:
                    self[subject] = logins


class MapfileManager(object):
    """
    """

    def __init__(self, argv=None, logger=None):
        """
        """
        self.argv = argv

        self.parse_command_line(argv)
        self.read_config_file()
        self.merge_options()

        self.parser.raise_unknown = True

        # Adding default source allows setting a default value using alternative syntax: --attribute_default = value
        for source in set(self.SOURCES + ["default"]):
            self.add_source_options(source)

        options, args = self.parser.parse_args(self.parser.largs)
        for option in vars(options):
            value = getattr(options, option)
            if value:
                setattr(self.options, option, value)

        self.merge_source_options()

        if self.options.DUMP_VARS:
            self.dump_vars()

        self.daemonize()
        if logger:
            self.logger = logger
        else:
            self.setup_logging()

        self.invalid_x509_subject_regexp = re.compile(r"(?<!\\)(?:\\{2})*\"")

    def parse_command_line(self, argv=None):
        """
        """
        if argv is None:
            argv = sys.argv

        # parse command line options
        self.parser = PassThroughOptionParser(
            version="lgmm " + __version__, description="LIGO Grid Mapfile Manager"
        )

        self.parser.add_option(
            "-c",
            "--config",
            dest="CONFIG_FILE_PATH",
            default="/etc/lgmm/lgmm_config.py",
            help="path to config file [default is /etc/lgmm/lgmm_config.py]",
        )

        self.parser.add_option(
            "--dump-vars",
            action="store_true",
            dest="DUMP_VARS",
            default=None,
            help="Dump configuration and exit",
        )

        self.parser.add_option(
            "--sources",
            dest="SOURCES",
            type="string",
            action="callback",
            callback=self._parse_list,
            default=None,
            help="Comma separated list of additional sources. [default is default]",
        )
        group = OptionGroup(self.parser, "Gridmapfile Input Options")

        group.add_option(
            "-b",
            "--blacklist",
            dest="BLACKLIST",
            default=None,
            help="grid-mapfile blacklist file [default is NONE]",
        )
        group.add_option(
            "-w",
            "--whitelist",
            dest="WHITELIST",
            default=None,
            help="grid-mapfile whitelist in grid-mapfile format [default is None]",
        )
        group.add_option(
            "-i",
            "--infile",
            dest="INFILE",
            default=None,
            help="File from which to read existing mappings [default is /etc/grid-security/grid-mapfile]",
        )
        self.parser.add_option_group(group)

        group = OptionGroup(self.parser, "Gridmapfile Output Options")
        group.add_option(
            "-g",
            "--grid-mapfile",
            "-o",
            "--outfile",
            dest="OUTFILE",
            default=None,
            help="File to which to write mappings [default is /etc/grid-security/grid-mapfile]",
        )
        group.add_option(
            "--sort",
            dest="SORT",
            default=None,
            help="sorting callable for sorting grid-mapfile [default is None]",
        )
        group.add_option(
            "--no-local-accounts",
            action="store_true",
            dest="NO_LOCAL_ACCOUNTS",
            default=None,
            help="Do not add mappings to local accounts and only include certificates subjects [default is FALSE]",
        )
        group.add_option(
            "--local-accounts",
            action="store_false",
            dest="NO_LOCAL_ACCOUNTS",
            default=None,
            help=SUPPRESS_HELP,
        )
        group.add_option(
            "-f",
            "--force",
            action="store_true",
            dest="FORCE",
            default=None,
            help="force new logins into new grid-mapfile [default is False]",
        )
        group.add_option(
            "--no-force", action="store_false", dest="FORCE", help=SUPPRESS_HELP
        )
        group.add_option(
            "--dry-run",
            action="store_true",
            dest="DRY_RUN",
            default=None,
            help="do not write any output [default is False]",
        )
        group.add_option(
            "--no-dry-run", action="store_false", dest="DRY_RUN", help=SUPPRESS_HELP
        )

        self.parser.add_option_group(group)

        group = OptionGroup(self.parser, "Logging Options")
        group.add_option(
            "--log-file",
            dest="LOGFILE",
            default=None,
            help="path to logfile [default is /var/log/lgmm/lgmm.log] ",
        )
        group.add_option(
            "--log-level",
            dest="LOG_LEVEL",
            type="string",
            action="callback",
            callback=self._parse_log_level,
            default=None,
            help="one of DEBUG, INFO, WARN, ERROR, CRITICAL, FATAL [default is INFO]",
        )
        self.parser.add_option_group(group)

        group = OptionGroup(self.parser, "Daemon Options")
        group.add_option(
            "-d",
            "--daemon",
            action="store_true",
            dest="DAEMON",
            default=None,
            help="detach and become daemon [default is False]",
        )
        group.add_option(
            "--pid-file",
            dest="PIDFILE",
            default=None,
            help="path to pidfile [default is /var/run/lgmm/lgmm.pid]",
        )
        group.add_option(
            "--sleep-time",
            dest="SLEEP_TIME",
            type="int",
            default=None,
            help="sleep time between processing cycles [default is 300 seconds]",
        )
        group.add_option(
            "--random-delay",
            action="store_true",
            dest="RANDOM_DELAY",
            default=None,
            help="if daemonized, delay startup by random time up to sleep time to load balance LDAP server [default is enabled, use --no-random-delay to disable]",
        )
        group.add_option(
            "--no-random-delay",
            action="store_false",
            dest="RANDOM_DELAY",
            default=None,
            help=SUPPRESS_HELP,
        )
        group.add_option(
            "--user",
            dest="USER",
            default=None,
            help="user to run as when daemon [default is nobody]",
        )
        self.parser.add_option_group(group)

        self.add_source_options()
        (options, args) = self.parser.parse_args(argv[1:])

        self.options = options
        self.args = args

    def add_source_options(self, source=None):
        def _args(sargs, largs, source):
            if source:
                return [attr_join(a, source) for a in as_list(largs)]
            return as_list(sargs) + as_list(largs)

        group = OptionGroup(
            self.parser,
            "LDAP Host Options",
            "Add source specific options using --argument_source=...",
        )

        group.add_option(
            *_args("-U", "--ldap-urls", source),
            type="string",
            dest=attr_join("LDAP_URLS", source, add_default=False),
            action="callback",
            callback=self._parse_ldap_urls,
            default=None,
            help="list of LDAP URLs separated by ',' [default is [ 'ldaps://ldap.ligo.org', 'ldaps://ldap.phys.uwm.edu' ]]"
        )

        group.add_option(
            *_args("-B", "--ldap-base", source),
            dest=attr_join("LDAP_BASE", source, add_default=False),
            default=None,
            help="LDAP search base [default is dc=ligo,dc=org]"
        )

        group.add_option(
            *_args("-S", "--ldap-scope", source),
            dest=attr_join("LDAP_SCOPE", source, add_default=False),
            default=None,
            help='LDAP search scope, can be one of "base", "onelevel" or "subtree" [default is "subtree"]'
        )

        group.add_option(
            *_args([], "--site-attr", source),
            dest=attr_join("SITE_ATTR", source, add_default=False),
            default=None,
            help="Login attribute 'site' option. {site} is replaced by SITE argument. [default is 'x-ligo-uid-{site}']"
        )

        self.parser.add_option_group(group)

        group = OptionGroup(
            self.parser,
            "LDAP Search Options",
            "Add source specific options using --argument_source=...",
        )

        group.add_option(
            *_args([], "--groups", source),
            type="string",
            dest=attr_join("GROUPS", source, add_default=False),
            action="callback",
            callback=self._parse_groups,
            default=None,
            help="required membership groups separated by ','. If specified multiple times each set of required groups are combined in OR fashion. [default is Communities:LSCVirgoLIGOGroupMembers]"
        )

        group.add_option(
            *_args(["-m", "-M"], "--membership-attribute", source),
            dest=attr_join("MEMBER_ATTR", source, add_default=False),
            default=None,
            help="LDAP attribute indicating group membership [default is isMemberOf]"
        )

        group.add_option(
            *_args("-L", "--login-attribute", source),
            dest=attr_join("LOGIN_ATTR", source, add_default=False),
            default=None,
            help="LDAP attribute indicating local logins [default is x-LIGO-uid]"
        )

        group.add_option(
            *_args([], "--subject-attribute", source),
            dest=attr_join("SUBJECT_ATTR", source, add_default=False),
            default=None,
            help="LDAP attribute indicating grid subject [default is gridX509subject]"
        )

        group.add_option(
            *_args([], "--filter", source),
            dest=attr_join("FILTER", source, add_default=False),
            default="",
            help="custom LDAP filter in standard query format e.g '(attribute=value)' [default is '']"
        )
        self.parser.add_option_group(group)

        group = OptionGroup(
            self.parser,
            "Additional Processing Options",
            "Add source specific options using --argument_source=...",
        )

        group.add_option(
            *_args("-s", "--site", source),
            dest=attr_join("SITE", source, add_default=False),
            default=None,
            help="site for which to generate grid-mapfile [default is None]"
        )

        group.add_option(
            *_args([], "--login-whitelist", source),
            dest=attr_join("LOGIN_WHITELIST", source, add_default=False),
            type="string",
            action="callback",
            callback=self._parse_list,
            default=None,
            help="Comma separated list of allowed logins. [default is None]"
        )

        group.add_option(
            *_args([], "--username-policy", source),
            dest=attr_join("USERNAME_POLICY", source, add_default=False),
            help="Policy on how to order site and global usernames, in the case that both exist. "
            'Acceptable values are {} [default is "SITE_ONLY"]'.format(
                ", ".join('"%s"' % (s,) for s in USERNAME_POLICIES)
            )
        )

        group.add_option(
            *_args([], "--login-groups-map", source),
            dest=attr_join("LOGIN_GROUPS_MAP", source, add_default=False),
            type="string",
            action="callback",
            callback=self._parse_mapping,
            default=None,
            help="Map local logins to groups. Each entry takes the form login[,login]:group[,group]... where a user is required to be in all groups in the list. [Default None]"
        )

        group.add_option(
            *_args([], "--subject-transform", source),
            dest=attr_join("SUBJECT_TRANSFORM", source, add_default=False),
            help="Whether to transform subject dn from  LDAP to grid format. "
            'Acceptable values are {} [default is "AUTO"]'.format(
                ", ".join('"%s"' % (s,) for s in SUBJECT_TRANSFORMS)
            )
        )

        group.add_option(
            *_args("-O", "--overrides", source),
            dest=attr_join("OVERRIDES", source, add_default=False),
            type="string",
            action="callback",
            callback=self._parse_overrides,
            default=None,
            help="List of override rules [default is None]"
        )
        self.parser.add_option_group(group)

    def _parse_ldap_urls(self, option, opt, value, parser):
        """
        """
        # TODO do more validation here
        setattr(parser.values, option.dest, value.split(","))

    def _parse_groups(self, option, opt, value, parser):
        """
        """
        # TODO do more validation here
        current_groups = getattr(parser.values, option.dest) or []
        current_groups.append(value.split(","))
        setattr(parser.values, option.dest, current_groups)

    def _parse_log_level(self, option, opt, value, parser):
        """
        """
        level = eval("logging." + value)

        # TODO do more validattion here
        setattr(parser.values, option.dest, level)

    def _parse_list(self, option, opt, value, parser):
        """
        """
        setattr(parser.values, option.dest, value.split(","))

    def _parse_mapping(self, option, opt, value, parser):
        """
        """
        current_maps = getattr(parser.values, option.dest) or []
        login, groups = value.split(":", 1)
        current_maps.append((login.split(","), groups.split(",")))
        setattr(parser.values, option.dest, current_maps)

    def _parse_overrides(self, option, opt, value, parser):
        """
        """
        overrides = getattr(parser.values, option.dest) or []
        overrides.extend(as_list(eval(value)))
        setattr(parser.values, option.dest, overrides)

    def read_config_file(self):
        """
        """
        if not os.path.isfile(self.options.CONFIG_FILE_PATH):
            msg = "Unable to import config file %s" % (self.options.CONFIG_FILE_PATH)
            raise RuntimeError(msg)

        try:
            spec = importlib_util.spec_from_file_location(
                "config", self.options.CONFIG_FILE_PATH
            )
            self.config = importlib_util.module_from_spec(spec)
            spec.loader.exec_module(self.config)
        except IOError as e:
            msg = "Unable to import config file %s: %s" % (
                self.options.configFilePath,
                e,
            )
            raise RuntimeError(msg)

    def merge_options(self):
        """
        """
        config = self.config
        options = self.options

        # command line has precedance over config file
        self.BLACKLIST = (
            options.BLACKLIST
            if options.BLACKLIST
            else config.BLACKLIST
            if hasattr(config, "BLACKLIST")
            else None
        )
        self.DAEMON = (
            options.DAEMON
            if (options.DAEMON is True or options.DAEMON is False)
            else config.DAEMON
            if hasattr(config, "DAEMON")
            else False
        )
        self.DRY_RUN = (
            options.DRY_RUN
            if (options.DRY_RUN is True or options.DRY_RUN is False)
            else config.DRY_RUN
            if hasattr(config, "DRY_RUN")
            else False
        )
        self.FORCE = (
            options.FORCE
            if (options.FORCE is True or options.FORCE is False)
            else config.FORCE
            if hasattr(config, "FORCE")
            else False
        )
        self.INFILE = (
            options.INFILE
            if options.INFILE
            else config.INFILE
            if hasattr(config, "INFILE")
            else config.GRID_MAPFILE
            if hasattr(config, "GRID_MAPFILE")
            else "/etc/grid-security/grid-mapfile"
        )
        self.PIDFILE = (
            options.PIDFILE
            if options.PIDFILE
            else config.PIDFILE
            if hasattr(config, "PIDFILE")
            else "/var/run/lgmm/lgmm.pid"
        )
        self.LOGFILE = (
            options.LOGFILE
            if options.LOGFILE
            else config.LOGFILE
            if hasattr(config, "LOGFILE")
            else "/var/log/lgmm/lgmm.log"
        )
        self.LOG_LEVEL = (
            options.LOG_LEVEL
            if options.LOG_LEVEL
            else config.LOG_LEVEL
            if hasattr(config, "LOG_LEVEL")
            else logging.INFO
        )
        self.NO_LOCAL_ACCOUNTS = (
            options.NO_LOCAL_ACCOUNTS
            if (options.NO_LOCAL_ACCOUNTS is True or options.NO_LOCAL_ACCOUNTS is False)
            else config.NO_LOCAL_ACCOUNTS
            if hasattr(config, "NO_LOCAL_ACCOUNTS")
            else False
        )
        self.OUTFILE = (
            options.OUTFILE
            if options.OUTFILE
            else config.OUTFILE
            if hasattr(config, "OUTFILE")
            else config.GRID_MAPFILE
            if hasattr(config, "GRID_MAPFILE")
            else "/etc/grid-security/grid-mapfile"
        )
        self.SLEEP_TIME = (
            options.SLEEP_TIME
            if options.SLEEP_TIME
            else config.SLEEP_TIME
            if hasattr(config, "SLEEP_TIME")
            else 300
        )
        self.RANDOM_DELAY = (
            options.RANDOM_DELAY
            if (options.RANDOM_DELAY is True or options.RANDOM_DELAY is False)
            else config.RANDOM_DELAY
            if hasattr(config, "RANDOM_DELAY")
            else True
        )
        self.SORT = (
            options.SORT
            if options.SORT
            else config.SORT
            if hasattr(config, "SORT")
            else None
        )
        self.USER = (
            options.USER
            if options.USER
            else config.USER
            if hasattr(config, "USER")
            else "nobody"
        )
        self.WHITELIST = (
            options.WHITELIST
            if options.WHITELIST
            else config.WHITELIST
            if hasattr(config, "WHITELIST")
            else None
        )

        self.SOURCES = (
            options.SOURCES
            if options.SOURCES
            else config.SOURCES
            if hasattr(config, "SOURCES")
            else ["default"]
        )

    def dump_vars(self):
        self.LOG_LEVEL = logging.getLevelName(self.LOG_LEVEL)

        keys = set(vars(self)) - set(
            ["args", "argv", "config", "options", "parser", "source_args"]
        )

        for key in sorted(keys):
            print("%s = %s" % (key, pformat(getattr(self, key, None), indent=4)))

        sys.exit()

    def merge_source_options(self):
        """
        """
        config = self.config
        options = self.options

        self.source_args = {
            "LDAP_URLS": ["ldaps://ldap.ligo.org", "ldaps://ldap.phys.uwm.edu"],
            "LDAP_BASE": "dc=ligo,dc=org",
            "LDAP_SCOPE": "subtree",
            "USERNAME_POLICY": "SITE_ONLY",
            "GROUPS": ["Communities:LSCVirgoLIGOGroupMembers"],
            "MEMBER_ATTR": "isMemberOf",
            "LOGIN_ATTR": "x-LIGO-uid",
            "SUBJECT_ATTR": "gridX509subject",
            "LOGIN_WHITELIST": None,
            "SITE": None,
            "SITE_ATTR": "x-ligo-ldg-{site}",
            "FILTER": None,
            "OVERRIDES": None,
            "LOGIN_GROUPS_MAP": None,
            "SUBJECT_TRANSFORM": "AUTO",
        }

        # Set default options
        for arg, default in list(self.source_args.items()):
            setattr(
                self,
                arg,
                (
                    getattr(options, arg, None)
                    or getattr(config, attr_join(arg, "default"), None)
                    or getattr(config, arg, None)
                    or default
                ),
            )

            # Delete the config attributes ATTR_default, which have just been converted to just ATTR
            if hasattr(config, attr_join(arg, "default")):
                delattr(config, attr_join(arg, "default"))

        self.verify_options("USERNAME_POLICY", USERNAME_POLICIES)
        self.verify_options("SUBJECT_TRANSFORM", SUBJECT_TRANSFORMS)
        self.fix_groups()
        self.get_ldap_scope()

        for source in self.SOURCES:
            variable_set = False
            for arg, default in list(self.source_args.items()):
                value = getattr(options, attr_join(arg, source), None) or getattr(
                    config, attr_join(arg, source), None
                )
                if value:
                    setattr(self, attr_join(arg, source), value)
                    variable_set = True

            if source != "default" and not variable_set:
                raise RuntimeError(
                    "No specific variables found for source '%s'" % source
                )

            self.verify_options("USERNAME_POLICY", USERNAME_POLICIES, source)
            self.verify_options("SUBJECT_TRANSFORM", SUBJECT_TRANSFORMS, source)
            self.fix_groups(source)
            self.get_ldap_scope(source)

    def verify_options(self, config, options, source="default"):
        value = self.get_config(config, source).upper()
        if value not in options:
            raise RuntimeError(
                "Invalid %s value '%s', must be one of %s."
                % (config, value, ", ".join(options))
            )
        if source == "default":
            self.set_config(config, value)
        else:
            self.set_config(config, value, source)

    def get_ldap_scope(self, source=None):
        if source and not hasattr(self, attr_join("LDAP_SCOPE", source)):
            return

        scope = self.get_config("LDAP_SCOPE", source)
        for scope_name, scope_value in LDAP_SCOPES.items():
            if scope_name.startswith(scope.lower()):
                self.set_config("LDAP_SCOPE", scope_name, source)
                break
        else:
            raise RuntimeError("Invalid LDAP Scope '%s'" % (scope,))

    def fix_groups(self, source=None):
        if source and not hasattr(self, attr_join("GROUPS", source)):
            return

        groups = self.get_config("GROUPS", source)
        for group in groups:
            if isinstance(group, (list, tuple)):
                for group in groups:
                    if isinstance(group, str):
                        raise RuntimeError(
                            "When using advanced group selection all groups must be within a sublist"
                        )
                break
        else:
            self.set_config("GROUPS", [groups], source)

    def get_config(self, key, source=None):
        if source:
            if source != "default" and source not in self.SOURCES:
                raise RuntimeError("Unknown source '{}'".format(source))

            return getattr(self, attr_join(key, source), getattr(self, key))

        return getattr(self, key)

    def set_config(self, key, value, source=None):
        if source:
            return setattr(self, attr_join(key, source), value)
        setattr(self, key, value)

    def remove_pid(self):
        """
        """
        try:
            os.unlink(self.PIDFILE)
        except OSError:
            pass

    def sig_term_handler(self, signum, frame):
        """
        Handle SIGTERM signals

        @param signum: signal number; see the documentation for the
                signal module

        @param frame: current stack frame; see the documentation for
                the signal module

        @return: None
        """
        self.done = True

    def sig_int_handler(self, signum, frame):
        """
        Handle SIGINT signals

        @param signum: signal number; see the documentation for the
                signal module

        @param frame: current stack frame; see the documentation for
                the signal module

        @return: None
        """
        self.done = True

    def daemonize(self):
        """
        """
        if not self.DAEMON:
            return

        # sanitize the path
        os.environ["PATH"] = "/bin:/usr/bin:/sbin:/usr/sbin"

        # exit if pid file exists and the PID it contains is
        # a valid PID
        if os.path.exists(self.PIDFILE):
            with open(self.PIDFILE, "r") as f:
                line = f.readline()
                pid = int(line.strip())

                try:
                    os.kill(pid, 0)

                    # if no exception than the process is running
                    # so we must stop now
                    msg = "%s pid file %s exists so exiting" % (
                        self.argv[0],
                        self.PIDFILE,
                    )

                    raise RuntimeError(msg)
                except OSError:
                    # no such process so the PID in the file
                    # is invalid and we should clean it up first
                    try:
                        os.unlink(self.PIDFILE)
                    except OSError:
                        pass

        # first fork on way to becoming a daemon
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                os._exit(0)
        except OSError as e:
            msg = "%s fork #1 failed: %d (%s)" % (self.argv[0], e.errno, e.strerror)
            raise RuntimeError(msg)

        # become the session leader on the new session
        # and process group leader of the new process group
        os.setsid()

        # second fork on way to becoming a daemon
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                os._exit(0)
        except OSError as e:
            msg = "%s fork #2 failed: %d (%s)" % (self.argv[0], e.errno, e.strerror)
            raise RuntimeError(msg)

        # change to the root directory
        try:
            os.chdir("/")
        except Exception as e:
            msg = "Failed to chdir to / : %s" % e
            raise RuntimeError(msg)

        # make sure we do not inherit any strange umask
        os.umask(0)

        # close all open file descriptions
        import resource

        maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
        if maxfd == resource.RLIM_INFINITY:
            # maxfd = MAXFD  <-- MAXFD is not defined!!!
            pass

        # Iterate through and close all file descriptors.
        for fd in range(0, maxfd):
            try:
                os.close(fd)
            except OSError:  # ERROR, fd wasn't open to begin with (ignored)
                pass

        # Redirect the standard I/O file descriptors to the
        # specified file.  Since the daemon has no controlling
        # terminal, most daemons redirect stdin, stdout, and
        # stderr to /dev/null.  This is done to prevent
        # side-effects from reads and writes to the standard
        # I/O file descriptors.

        # This call to open is guaranteed to return the lowest
        # file descriptor, which will be 0 (stdin), since it
        # was closed above.

        os.open("/dev/null", os.O_RDWR)  # standard input (0)

        # Duplicate standard input to standard output and standard error.
        os.dup2(0, 1)  # standard output (1)
        os.dup2(0, 2)  # standard error (2)

        # create pid file and write PID into it
        try:
            with os.fdopen(
                os.open(self.PIDFILE, os.O_WRONLY | os.O_CREAT, 0o644), "w"
            ) as fh:
                msg = "%d" % os.getpid()
                fh.write(msg)
        except OSError:
            msg = "%s unable to create pid file %s" % (self.argv[0], self.PIDFILE)
            raise RuntimeError(msg)

        # make sure the user we will run as can delete the PID file
        uid = pwd.getpwnam(self.USER).pw_uid
        gid = pwd.getpwnam(self.USER).pw_gid

        pid_dir, pid_file = os.path.split(self.PIDFILE)

        try:
            os.chown(pid_dir, uid, gid)
            os.chown(self.PIDFILE, uid, gid)
            os.chmod(pid_dir, 0o755)
            os.chmod(self.PIDFILE, 0o644)
        except OSError as e:
            msg = "%s unable to manage pid file %s: %s" % (
                self.argv[0],
                self.PIDFILE,
                e,
            )
            raise RuntimeError(msg)

        # make sure the user we will run as can write to log file
        try:
            open(self.LOGFILE, "a").close()
            os.chown(self.LOGFILE, uid, gid)
            os.chmod(self.LOGFILE, 0o644)
        except (IOError, OSError) as e:
            msg = "%s unable to manage log file %s: %s" % (
                self.argv[0],
                self.LOGFILE,
                e,
            )
            raise RuntimeError(msg)

        # drop privileges
        os.setgid(gid)
        os.setuid(uid)

        # register function to remove pid file when exiting nicely
        atexit.register(self.remove_pid)

        # setup to catch SIGTERM, SIGHUP, and SIGINT
        signal.signal(signal.SIGTERM, self.sig_term_handler)
        signal.signal(signal.SIGINT, self.sig_int_handler)

    def setup_logging(self):
        """
        """
        # set up logging for the root
        logger = logging.getLogger("lgmm")

        # using logrotate for log rotation so just use the basic handler
        handler = logging.handlers.WatchedFileHandler(self.LOGFILE, "a")
        formatter = logging.Formatter("%(asctime)s %(levelname)-6s %(message)s")
        handler.setFormatter(formatter)

        logger.addHandler(handler)
        logger.setLevel(self.LOG_LEVEL)

        self.logger = logger
        self.logger.info("%s has started" % self.argv[0])

    def random_delay(self):
        """
        """
        random_delaytime = random.randint(0, self.SLEEP_TIME)
        self.logger.info(
            "Delaying startup by %d seconds to load balance LDAP queries"
            % random_delaytime
        )
        time.sleep(random_delaytime)
        self.logger.info("Awake!")

    def sleep(self):
        """
        """
        self.logger.info("Going to sleep for %d seconds..." % self.SLEEP_TIME)
        time.sleep(self.SLEEP_TIME)
        self.logger.info("Awake!")

    def fold_new_mapping(self, existing, white, black, central):
        """
        """
        logger = self.logger

        # prepare dictionary of all existing logins
        existing_logins = {}
        for subject, logins in list(existing.items()):
            for login in logins:
                logger.debug("%s is an existing login" % login)
                existing_logins[login] = 1

        # prepare dictionary of all blacklisted logins
        black_logins = {}
        for subject, logins in list(black.items()):
            for login in logins:
                logger.debug("%s is a blacklisted login" % login)
                black_logins[login] = 1

        # copy central in unless login does not exist, or is
        # blacklisted, unless we are being forced
        new = Mapping()
        rejections = {}
        for subject, logins in list(central.items()):
            invalid_subject_strings = self.invalid_x509_subject_regexp.findall(subject)
            if invalid_subject_strings:
                logger.info(
                    "Subject %s rejected as it contains illegal or unescaped characters: %s"
                    % (subject, " ".join(invalid_subject_strings))
                )
                continue

            if subject in black:
                logger.info("Subject %s is blacklisted" % subject)
                rejections[subject] = 1
                continue
            new_logins = []

            for login in logins:
                if login in black_logins:
                    logger.info("Login %s is blacklisted" % login)
                    logger.info("Not mapping %s to %s" % (subject, login))
                    rejections[login] = 1
                    continue
                elif login not in existing_logins:
                    if not self.FORCE:
                        logger.info("Login %s does not already exist" % login)
                        logger.info("Not mapping %s to %s" % (subject, login))
                        rejections[login] = 1
                        continue
                    else:
                        msg = "Login %s is being forced" % login
                        logger.info(msg)
                        print(msg)
                        msg = "Mapping %s to %s" % (subject, login)
                        logger.info(msg)
                        print(msg)
                        new_logins.append(login)
                else:
                    new_logins.append(login)

            if new_logins:
                logger.debug("Mapping %s to %s" % (subject, new_logins))
                new[subject] = new_logins

        # now add in the whitelist and give ordering
        # preference to the whitelist
        for subject, logins in list(white.items()):
            if subject in rejections:
                logger.info("Subject %s previously was rejected" % subject)
                logger.info("Adding %s since whitelisted" % subject)

            for login in logins:
                if login in rejections:
                    logger.info("Login %s previously was rejected" % login)
                    logger.info("Adding %s since whitelisted" % login)

            if subject in new:
                new_logins = logins
                new_logins.extend(new[subject])
                new_logins = order_preserve_uniq(new_logins)
            else:
                new_logins = logins

            logger.debug("Whitelisting mapping of %s to %s" % (subject, new_logins))
            new[subject] = new_logins

        return new

    def run(self):
        """
        """
        logger = self.logger

        self.done = False

        # loop until marked done either after one iteration
        # when not running in daemon mode or after catching
        # signal

        if self.DAEMON and self.RANDOM_DELAY:
            self.random_delay()

        central = CentralMapping(self, fetch_now=False)

        while not self.done:
            try:
                # load existing grid-mapfile
                existing = ExistingMapping(self.INFILE, required=False)
                logger.info(
                    "Found %d mappings in %s"
                    % (len(list(existing.keys())), self.INFILE)
                )

                # load whitelist if configured
                if self.WHITELIST:
                    white = ExistingMapping(self.WHITELIST)
                    logger.info(
                        "Found %d mappings in whitelist %s"
                        % (len(list(white.keys())), self.WHITELIST)
                    )
                else:
                    white = Mapping()
                    logger.info("No whitelist configured")

                # load blacklist if configured
                if self.BLACKLIST:
                    black = BlackListMapping(self.BLACKLIST)
                    logger.info(
                        "Found %d lines in blacklist %s"
                        % (len(list(black.keys())), self.BLACKLIST)
                    )
                else:
                    black = Mapping()
                    logger.info("No blacklist configured")

                # query LDAP for mappings
                central.refresh()
                logger.info("Found %d mappings in LDAP" % len(list(central.keys())))

                # use existing, white and black lists, and
                # central mappings to create a new mapping
                new = self.fold_new_mapping(existing, white, black, central)

                logger.info("Preparing to write grid-mapfile to %s" % self.OUTFILE)
                new.persist_to_file(
                    self.OUTFILE,
                    logger,
                    sort=self.SORT,
                    no_local_accounts=self.NO_LOCAL_ACCOUNTS,
                    dry_run=self.DRY_RUN,
                )

            except RuntimeError as e:
                logger.error("Error: %s" % e)

            if self.DAEMON:
                self.sleep()

            else:
                self.done = True

        self.logger.info("%s is exiting" % self.argv[0])


def main(argv=None):
    """
    """
    if argv is None:
        argv = sys.argv
    manager = MapfileManager(argv=argv)
    return manager.run()


if __name__ == "__main__":
    sys.exit(main(sys.argv))
