# -*- coding: utf-8 -*-

import tempfile
import unittest

import ldap3

import pytest

import lgmm

from .common import (
    LGMM_CONFIG,
    MockLIGOLDAPWithManager,
    logger,
)
from .data import (
    data,
    directory,
)


class MappingTests(unittest.TestCase):
    def test_required_arguments(self):
        mapping = lgmm.Mapping()
        with self.assertRaises(TypeError):
            mapping.persist_to_file()

        with tempfile.NamedTemporaryFile() as tmp:
            with self.assertRaises(TypeError):
                mapping.persist_to_file(tmp.name)

        with tempfile.NamedTemporaryFile() as tmp:
            mapping.persist_to_file(tmp.name, logger)

    def test_mapping_output(self):
        dn = data["einstein_dn"]
        usernames = data["einstein_usernames"]

        mapping = lgmm.Mapping()
        mapping[dn] = usernames

        with tempfile.NamedTemporaryFile() as tmp:
            mapping.persist_to_file(tmp.name, logger)
            tmp.flush()

            with open(tmp.name, "r") as tmp:
                output = tmp.read().splitlines()

            self.assertEqual(output, [data["einstein_mapfile_entry"]])

    def test_mapping_output_no_local(self):
        dn = data["einstein_dn"]
        usernames = data["einstein_usernames"]

        mapping = lgmm.Mapping()
        mapping[dn] = usernames

        with tempfile.NamedTemporaryFile() as tmp:
            mapping.persist_to_file(tmp.name, logger, no_local_accounts=True)
            tmp.flush()

            with open(tmp.name, "r") as tmp:
                output = tmp.read()

            self.assertEqual(output, data["einstein_dn_quoted"] + "\n")

    def test_mapping_output_dry_run(self):
        dn = data["einstein_dn"]
        usernames = data["einstein_usernames"]

        mapping = lgmm.Mapping()
        mapping[dn] = usernames

        with tempfile.NamedTemporaryFile() as tmp:
            mapping.persist_to_file(tmp.name, logger, dry_run=True)
            tmp.flush()

            with open(tmp.name, "r") as tmp:
                output = tmp.read()

            expected = ""
            self.assertEqual(output, expected)

    def test_mapping_output_sort(self):
        dn1 = data["einstein_dn"]
        usernames1 = data["einstein_usernames"]

        dn2 = data["curie_dn"]
        usernames2 = data["curie_usernames"]

        mapping = lgmm.Mapping()
        mapping[dn1] = usernames1
        mapping[dn2] = usernames2

        def get_surname(line):
            return line.split('"', 2)[2].split(".", 1)[1]

        with tempfile.NamedTemporaryFile() as tmp:
            mapping.persist_to_file(tmp.name, logger, sort=get_surname)
            tmp.flush()

            with open(tmp.name, "r") as tmp:
                output = tmp.read().splitlines()

            expected = [data["curie_mapfile_entry"], data["einstein_mapfile_entry"]]
            self.assertEqual(output, expected)


class FileMappingTests(unittest.TestCase):
    def test_required_arguments(self):
        with self.assertRaises(TypeError):
            lgmm.FileMapping()

        with tempfile.NamedTemporaryFile() as tmp:
            mapping = lgmm.FileMapping(tmp.name)

            mapping.parse()

    def test_parseable_file(self):
        with tempfile.NamedTemporaryFile() as tmp:
            with open(tmp.name, "w") as tmp:
                # Normal entry
                tmp.write(data["einstein_mapfile_entry"] + "\n")
                # No usernames
                tmp.write(data["curie_dn_quoted"] + "\n")
                # Commented line
                tmp.write("#" + data["curie_mapfile_entry"] + "\n")
                # Blank line
                tmp.write("\n")

            mapping = lgmm.FileMapping(tmp.name)
            mapping.parse()

        self.assertTrue(mapping.parsed)
        self.assertEqual(2, len(mapping))

        self.assertIn(data["einstein_dn"], mapping)
        self.assertEqual(mapping[data["einstein_dn"]], data["einstein_usernames"])

        self.assertIn(data["curie_dn"], mapping)
        self.assertEqual(mapping[data["curie_dn"]], [])

    def test_unparseable_file(self):
        with tempfile.NamedTemporaryFile() as tmp:
            with open(tmp.name, "w") as tmp:
                # Missing first " character
                tmp.write(data["einstein_mapfile_entry"][1:] + "\n")

            mapping = lgmm.FileMapping(tmp.name)
            with self.assertRaises(RuntimeError):
                mapping.parse()


class ExistingMappingTests(unittest.TestCase):
    def test_nonexisting_file(self):
        with tempfile.NamedTemporaryFile() as tmp:
            with open(tmp.name, "w") as tmp:
                tmp.write(data["einstein_mapfile_entry"] + "\n")

            mapping = lgmm.ExistingMapping(tmp.name)
            mapping.parse()

        mapping = lgmm.ExistingMapping(tmp.name, required=False)

        with self.assertRaises(IOError):
            mapping = lgmm.ExistingMapping(tmp.name)


class BlackListMappingTests(unittest.TestCase):
    def test_blacklist_mapping_file(self):
        with tempfile.NamedTemporaryFile() as tmp:
            with open(tmp.name, "w") as tmp:
                # normal entry
                tmp.write(data["einstein_mapfile_entry"] + "\n")
                # blacklisted usernames only
                tmp.write(data["curie_usernames"][0] + "\n")

            mapping = lgmm.BlackListMapping(tmp.name)

        self.assertTrue(mapping.parsed)
        self.assertEqual(2, len(mapping))

        self.assertIn(data["einstein_dn"], mapping)
        self.assertEqual(mapping[data["einstein_dn"]], data["einstein_usernames"])

        self.assertIn("1", mapping)
        self.assertEqual(mapping["1"], data["curie_usernames"])


@pytest.mark.skipif(
    ldap3.__version__ < "2.5.0",
    reason="ldap3 too old (< 2.5.0)",
)
class CentralManagerLIGOTests(MockLIGOLDAPWithManager):
    """
    Test CentralManager functionality using standard LIGO schema
    """

    directory = directory

    @unittest.skip("Does not check for failed LDAP server")
    def test_multiple_ldap_urls(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        manager.LDAP_URLS = ["ldaps://0.0.0.0/", "ldaps://ldap.ligo.org"]
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))

        self.assertIn(data["einstein_dn"], mapping)

    def test_ldap_base_and_scope(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        manager.LDAP_SCOPE = "onelevel"
        mapping = lgmm.CentralMapping(manager)
        self.assertEqual(0, len(mapping))

        manager.LDAP_BASE = "ou=people,dc=ligo,dc=org"
        mapping = lgmm.CentralMapping(manager)
        self.assertEqual(3, len(mapping))

        self.assertIn(data["einstein_dn"], mapping)

        manager.LDAP_BASE = "employeeNumber=1,ou=people,dc=ligo,dc=org"
        manager.LDAP_SCOPE = "base"
        mapping = lgmm.CentralMapping(manager)
        self.assertEqual(1, len(mapping))

        self.assertIn(data["einstein_dn"], mapping)

    def test_ldap_base_missing(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        manager.LDAP_BASE = "ou=peeple,dc=ligo,dc=org"
        manager.LDAP_SCOPE = "onelevel"
        with self.assertRaises(RuntimeError):
            lgmm.CentralMapping(manager)

    def test_all_ligovirgo_members(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))

        self.assertIn(data["einstein_dn"], mapping)
        self.assertEqual(mapping[data["einstein_dn"]], data["einstein_usernames"])

        self.assertIn(data["curie_dn"], mapping)
        self.assertEqual(mapping[data["curie_dn"]], data["curie_usernames"])

        self.assertIn(data["dirac_dn"], mapping)
        self.assertEqual(mapping[data["dirac_dn"]], data["dirac_usernames"])

    def test_all_ligovirgo_and_ldg_members(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        manager.GROUPS = (
            [
                "Communities:LSCVirgoLIGOGroupMembers",
                "Communities:LVC:LSC:LDG:LDGUsers",
            ],
        )
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(2, len(mapping))

        self.assertIn(data["einstein_dn"], mapping)
        self.assertEqual(mapping[data["einstein_dn"]], data["einstein_usernames"])

        self.assertIn(data["curie_dn"], mapping)
        self.assertEqual(mapping[data["curie_dn"]], data["curie_usernames"])

    def test_one_robot(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )

        manager.GROUPS = [["Communities:Robots"]]
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(1, len(mapping))

        self.assertEqual(mapping.get(data["robot_dn"]), data["robot_usernames"])

    def test_all_ldgusers_and_one_robot(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )

        manager.GROUPS = [
            [
                "Communities:LSCVirgoLIGOGroupMembers",
                "Communities:LVC:LSC:LDG:LDGUsers",
            ],
            ["Communities:Robots"],
        ]
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))

        self.assertEqual(mapping.get(data["robot_dn"]), data["robot_usernames"])

    def test_filter(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )

        manager.FILTER = "(x-LIGO-uid=albert.einstein)"
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(1, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), data["einstein_usernames"])

        manager.FILTER = "x-LIGO-uid=albert.einstein"
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(1, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), data["einstein_usernames"])

    def test_site(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )

        manager.SITE = "cdf"
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames_cdf"]
        )

    def test_login_whitelist(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )

        manager.LOGIN_WHITELIST = ["albert.einstein"]
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), ["albert.einstein"])
        self.assertEqual(mapping.get(data["curie_dn"]), [])
        self.assertEqual(mapping.get(data["dirac_dn"]), [])

    def test_username_policy(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        manager.SITE = "cdf"

        self.assertEqual(manager.USERNAME_POLICY, "SITE_ONLY")
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames_cdf"]
        )

        manager.USERNAME_POLICY = "SITE_FIRST"
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(
            mapping.get(data["einstein_dn"]),
            data["einstein_usernames_cdf"] + data["einstein_usernames"],
        )

        manager.USERNAME_POLICY = "GLOBAL_ONLY"
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), data["einstein_usernames"])

        manager.USERNAME_POLICY = "GLOBAL_FIRST"
        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]),
            data["einstein_usernames"] + data["einstein_usernames_cdf"],
        )

    def test_login_groups_map(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        manager.LOGIN_GROUPS_MAP = [
            ("cbc", ["Communities:LVC:LSC:LDG:Accounts:CBC:LDGAccountsCBCCBCUsers"])
        ]

        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames"] + ["cbc"]
        )

        manager.LOGIN_GROUPS_MAP = [
            ("cbc", ["Communities:LVC:LSC:LDG:Accounts:CBC:LDGAccountsCBCCBCUsers"]),
            (
                "xpipeline",
                [
                    "Communities:LVC:LSC:LDG:Accounts:Burst:LDGAccountsBurstXpipelineUsers"
                ],
            ),
        ]

        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames"] + ["cbc"]
        )
        self.assertEqual(
            mapping.get(data["curie_dn"]),
            data["curie_usernames"] + ["cbc", "xpipeline"],
        )

        manager.LOGIN_GROUPS_MAP = [
            (
                "cbcxpipeline",
                [
                    "Communities:LVC:LSC:LDG:Accounts:Burst:LDGAccountsBurstXpipelineUsers",
                    "Communities:LVC:LSC:LDG:Accounts:CBC:LDGAccountsCBCCBCUsers",
                ],
            )
        ]

        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), data["einstein_usernames"])
        self.assertEqual(
            mapping.get(data["curie_dn"]), data["curie_usernames"] + ["cbcxpipeline"]
        )

        manager.LOGIN_GROUPS_MAP = [
            ("cbc", ["Communities:LVC:LSC:LDG:Accounts:CBC:LDGAccountsCBCCBCUsers"]),
            (
                "cbc",
                ["Communities:LVC:LSC:LDG:Accounts:CBC:LDGAccountsCBCGSTLALCBCUsers"],
            ),
        ]

        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames"] + ["cbc"]
        )
        self.assertEqual(
            mapping.get(data["curie_dn"]), data["curie_usernames"] + ["cbc"]
        )
        self.assertEqual(
            mapping.get(data["dirac_dn"]), data["dirac_usernames"] + ["cbc"]
        )

        manager.LOGIN_GROUPS_MAP = [
            (
                ["gstlalcbc", "gstlalcbctest"],
                ["Communities:LVC:LSC:LDG:Accounts:CBC:LDGAccountsCBCGSTLALCBCUsers"],
            )
        ]

        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), data["einstein_usernames"])
        self.assertEqual(
            mapping.get(data["dirac_dn"]),
            data["dirac_usernames"] + ["gstlalcbc", "gstlalcbctest"],
        )

        manager.LOGIN_GROUPS_MAP = [
            (
                "cbc",
                [
                    "Communities:LVC:LSC:LDG:Accounts:CBC:LDGAccountsCBCCBCUsers",
                    "Communities:LVC:LSC:LDG:Accounts:CBC:LDGAccountsCBCCBCRotaUsers",
                ],
            )
        ]

        mapping = lgmm.CentralMapping(manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), data["einstein_usernames"])

    def test_overrides(self):
        self.manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )

        self.manager.OVERRIDES = [
            {
                "conditions": {"eduPersonPrincipalName": []},
                "logins_append": ["{eduPersonPrincipalName}"],
            }
        ]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]),
            data["einstein_usernames"] + ["albert.einstein@ligo.org"],
        )
        self.assertEqual(
            mapping.get(data["curie_dn"]),
            data["curie_usernames"] + ["marie.curie@ligo.org"],
        )
        self.assertEqual(
            mapping.get(data["dirac_dn"]),
            data["dirac_usernames"] + ["paul.dirac@ligo.org"],
        )

    def test_override_action_logins_append(self):
        self.manager.OVERRIDES = [{"conditions": {}, "logins_append": ["ligo"]}]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames"] + ["ligo"]
        )

    def test_override_action_logins_prepend(self):
        self.manager.OVERRIDES = [{"conditions": {}, "logins_prepend": ["ligo"]}]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]), ["ligo"] + data["einstein_usernames"]
        )

    def test_override_action_logins_set(self):
        self.manager.OVERRIDES = [{"conditions": {}, "logins_set": ["ligo"]}]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), ["ligo"])

    def test_override_action_logins_remove(self):
        self.manager.OVERRIDES = [{"conditions": {}, "logins_remove": ["albert"]}]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), ["albert.einstein"])
        self.assertEqual(mapping.get(data["curie_dn"]), data["curie_usernames"])

    def test_override_action_logins_remove_by_index(self):
        self.manager.OVERRIDES = [{"conditions": {}, "logins_remove": [0]}]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), ["albert"])
        self.assertEqual(mapping.get(data["curie_dn"]), [])

    def test_override_action_subjects_add(self):
        # add extra subject to all matching users
        self.manager.OVERRIDES = [{"conditions": {}, "subjects_add": ["/X509/LDGUSer"]}]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(4, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), data["einstein_usernames"])

        # lgmm dictates that certificate cannot be mapped to more than one account
        # May map to [albert.einstein, albert], or paul.dirac, depending on order
        self.assertLessEqual(len(mapping.get("/X509/LDGUSer")), 2)

    def test_override_action_subjects_set(self):
        # set subject on all matching usser
        self.manager.OVERRIDES = [{"conditions": {}, "subjects_set": ["/X509/LDGUSer"]}]
        mapping = lgmm.CentralMapping(self.manager)
        self.assertEqual(1, len(mapping))

        # lgmm dictates that certificate cannot be mapped to more than one account
        # May map to [albert.einstein, albert], or paul.dirac, depending on order
        self.assertLessEqual(len(mapping.get("/X509/LDGUSer")), 2)

    def test_override_action_subject_remove(self):
        self.manager.OVERRIDES = [
            {"conditions": {}, "subjects_remove": [data["einstein_dn"]]}
        ]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(2, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), None)

    def test_override_action_subject_remove_by_index(self):
        self.manager.OVERRIDES = [{"conditions": {}, "subjects_remove": [0]}]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(0, len(mapping))

    def test_override_action_logins_append_with_value(self):
        self.manager.OVERRIDES = [
            {
                "conditions": {"eduPersonPrincipalName": []},
                "logins_append": ["{eduPersonPrincipalName}"],
            }
        ]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]),
            data["einstein_usernames"] + ["albert.einstein@ligo.org"],
        )

    def test_override_action_logins_prepend_with_value(self):
        self.manager.OVERRIDES = [
            {
                "conditions": {"eduPersonPrincipalName": []},
                "logins_prepend": ["{eduPersonPrincipalName}"],
            }
        ]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]),
            ["albert.einstein@ligo.org"] + data["einstein_usernames"],
        )

    def test_override_no_conditions(self):
        self.manager.OVERRIDES = [{"conditions": {}, "logins_append": ["ligo"]}]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames"] + ["ligo"]
        )

    def test_override_one_condition_one_attribute(self):
        self.manager.GROUPS = []
        self.manager.OVERRIDES = [
            {"conditions": {"givenName": ["Albert"]}, "logins_append": ["ligo"]}
        ]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(5, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames"] + ["ligo"]
        )
        self.assertEqual(
            mapping.get(data["hofmann_dn"]), data["hofmann_usernames"] + ["ligo"]
        )

    def test_override_one_condition_two_attributes(self):
        self.manager.GROUPS = []
        self.manager.OVERRIDES = [
            {
                "conditions": {"givenName": ["Albert"], "sn": ["Einstein"]},
                "logins_append": ["ligo"],
            }
        ]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(5, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames"] + ["ligo"]
        )
        self.assertEqual(mapping.get(data["hofmann_dn"]), data["hofmann_usernames"])

    def test_override_two_conditions_one_attribute(self):
        self.manager.GROUPS = []
        self.manager.OVERRIDES = [
            {
                "conditions": {
                    "isMemberOf": [
                        "Communities:LSCVirgoLIGOGroupMembers",
                        "Communities:LVC:LSC:LDG:LDGUsers",
                    ]
                },
                "logins_append": ["ligo"],
            }
        ]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(5, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames"] + ["ligo"]
        )
        self.assertEqual(
            mapping.get(data["curie_dn"]), data["curie_usernames"] + ["ligo"]
        )
        self.assertEqual(mapping.get(data["dirac_dn"]), data["dirac_usernames"])

    def test_override_two_conditions_two_attributes(self):
        self.manager.GROUPS = []
        self.manager.OVERRIDES = [
            {
                "conditions": {
                    "givenName": ["Albert"],
                    "isMemberOf": [
                        "Communities:LSCVirgoLIGOGroupMembers",
                        "Communities:LVC:LSC:LDG:LDGUsers",
                    ],
                },
                "logins_append": ["ligo"],
            }
        ]
        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(5, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames"] + ["ligo"]
        )
        self.assertEqual(mapping.get(data["curie_dn"]), data["curie_usernames"])
        self.assertEqual(mapping.get(data["dirac_dn"]), data["dirac_usernames"])

    def test_override_negative_conditions(self):
        self.manager.GROUPS = []

    def test_override_negative_condition(self):
        self.manager.OVERRIDES = [
            {"conditions": {"givenName": ["!Albert"]}, "logins_append": ["not.albert"]}
        ]

        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(mapping.get(data["einstein_dn"]), data["einstein_usernames"])
        self.assertEqual(
            mapping.get(data["curie_dn"]), data["curie_usernames"] + ["not.albert"]
        )

    def test_override_multiple_rules(self):
        self.manager.OVERRIDES = [
            {"conditions": {"givenName": ["Albert"]}, "logins_append": ["albert"]},
            {"conditions": {"givenName": ["Paul"]}, "logins_append": ["paul"]},
        ]

        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames"] + ["albert"]
        )
        self.assertEqual(
            mapping.get(data["dirac_dn"]), data["dirac_usernames"] + ["paul"]
        )
        self.assertEqual(mapping.get(data["curie_dn"]), data["curie_usernames"])

    def test_override_regex_conditions(self):
        self.manager.OVERRIDES = [
            {
                "conditions": {"givenName": ["~^(Albert|Paul)$"]},
                "logins_append": ["albert.paul"],
            }
        ]

        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))
        self.assertEqual(
            mapping.get(data["einstein_dn"]),
            data["einstein_usernames"] + ["albert.paul"],
        )
        self.assertEqual(
            mapping.get(data["dirac_dn"]), data["dirac_usernames"] + ["albert.paul"]
        )
        self.assertEqual(mapping.get(data["curie_dn"]), data["curie_usernames"])


@pytest.mark.skipif(
    ldap3.__version__ < "2.5.0",
    reason="ldap3 too old (< 2.5.0)",
)
class CentralManagerLIGOAlternativeSchemaTests(MockLIGOLDAPWithManager):
    """
    Test CentralManager functionality using alternative schema
    """

    directory = directory

    def test_membership_and_login_attrs(self):
        self.manager.LDAP_BASE = "dc=ligo-alt,dc=org"
        self.manager.MEMBER_ATTR = "groupMembership"
        self.manager.LOGIN_ATTR = "uid"
        self.manager.SITE = "cardiff"
        self.manager.SITE_ATTR = "x-private-{site}-attr"

        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(3, len(mapping))

        self.assertEqual(
            mapping.get(data["einstein_dn"]), data["einstein_usernames_cdf"]
        )

    def test_x509subject_transform_auto(self):
        self.manager.LDAP_BASE = "dc=ligo-alt,dc=org"
        self.manager.MEMBER_ATTR = "groupMembership"
        self.manager.LOGIN_ATTR = "uid"
        self.manager.SUBJECT_ATTR = "gridX509subject_alt"
        self.manager.SUBJECT_TRANSFORM = "AUTO"

        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(1, len(mapping))

        self.assertEqual(mapping.get(data["einstein_dn"]), data["einstein_usernames"])

    def test_x509subject_transform_none(self):
        self.manager.LDAP_BASE = "dc=ligo-alt,dc=org"
        self.manager.MEMBER_ATTR = "groupMembership"
        self.manager.LOGIN_ATTR = "uid"
        self.manager.SUBJECT_ATTR = "gridX509subject_alt"
        self.manager.SUBJECT_TRANSFORM = "NONE"

        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(1, len(mapping))

        self.assertEqual(
            mapping.get(
                "cn=Albert Einstein albert.einstein@ligo.org,o=LIGO,c=US,dc=cilogon,dc=org"
            ),
            data["einstein_usernames"],
        )

    def test_x509subject_transform_force(self):
        self.manager.LDAP_BASE = "dc=ligo-alt,dc=org"
        self.manager.MEMBER_ATTR = "groupMembership"
        self.manager.LOGIN_ATTR = "uid"
        self.manager.SUBJECT_ATTR = "gridX509subject"
        self.manager.SUBJECT_TRANSFORM = "FORCE"

        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(len(mapping), 0)

    def test_x509subject_transform_auto_with_attribute_option_prior(self):
        self.manager.LDAP_BASE = "dc=ligo-alt,dc=org"
        self.manager.MEMBER_ATTR = "groupMembership"
        self.manager.LOGIN_ATTR = "uid"
        self.manager.SUBJECT_ATTR = "gridX509subject_alt_with_prior"
        self.manager.SUBJECT_TRANSFORM = "AUTO"

        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(0, len(mapping))

    def test_x509subject_transform_auto_with_attribute_option_scope(self):
        self.manager.LDAP_BASE = "dc=ligo-alt,dc=org"
        self.manager.MEMBER_ATTR = "groupMembership"
        self.manager.LOGIN_ATTR = "uid"
        self.manager.SUBJECT_ATTR = "gridX509subject_alt_with_scope"
        self.manager.SUBJECT_TRANSFORM = "AUTO"

        mapping = lgmm.CentralMapping(self.manager)

        self.assertEqual(1, len(mapping))

        self.assertEqual(mapping.get(data["einstein_dn"]), data["einstein_usernames"])
